(**
This file is part of the IE library

Copyright (C) El Ouraoui, Mayero, Mazza

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
COPYING file for more details.
*)

Require Export Sn_Axiomatic.
 

(** Ordre lexicographique sur les paires d'entiers *)
Inductive lex : (nat * nat) -> (nat * nat) -> Prop  :=
 | lt_L : forall s1 s2,  (fst s1) < (fst s2) -> lex s1 s2
 | lt_G : forall s1 s2, (fst s1) = (fst s2) -> (snd s1) < (snd s2)
                         -> lex s1 s2.

Notation "a <lex b" := (lex a b) (at level 70).


(* la premiere composante dans une définition pour rendre 
les choses plus lisibles*)
Definition f_cp t u := (eta ((t {0 := u })) ) + (eta u).
Definition alpha (t:term) (u:term) :=
   ((f_cp t u ),(num_occ t 0)).


(** Induction forte; preuve en cours *)

Axiom sn_ind :
  forall P : term -> term -> Prop,
  (forall t u:term,
    (forall t' u':term,
       (alpha t' u') <lex  (alpha t u) -> P t' u')
    -> (P t u) )
  -> forall t u:term, P t u .
(*
Theorem strong_induction:
   forall P : term -> term -> Prop,
  (forall t u:term,
    (forall t' u':term,
       (alpha t' u') <lex  (alpha t u) -> P t' u')
    -> (P t u) )
  -> forall t u:term, P t u .
Proof.
  intros P H t u.
  assert (forall t' u', (alpha t' u') <lex (alpha t u) -> P t u).
  induction alpha.
  intros k Hk.
  replace k with 0%nat.
  apply H.
 intros m Hm; contradict Hm.
 apply lt_n_0.
 generalize Hk; case k; trivial.
 intros m Hm; contradict Hm.
 apply le_not_lt.
 now auto with arith.
 intros k Hk.
 apply H.
 intros m Hm.
 apply IHn.
 apply lt_le_trans with (1:=Hm).
 now apply gt_S_le.
 apply H0.
 apply le_n.
Qed.  
  **)
                                                          
Lemma One_step_ls :
  forall t t' u, 
    (t [u]) -->ls (t' [u]) -> num_occ t' 0 < num_occ t 0.
Proof.  
  -intros t t' u. inv. clear H. revert H1. intros H. rewrite <- lift0 in H at 1. apply ls_equivalence in H.
   *induction H.
   +simpl. case Nat.eqb_spec. intros. unfold lift. rewrite (free_lift u (S i) i); omega.
    intros. omega.
   +simpl; apply IHlsa_d_ret.
   +simpl. omega.
   +simpl; auto with arith.
   +simpl; auto with arith.
   +simpl; auto with arith.
Qed.

Lemma subst_var_0_eta :
  forall u,forall i,
    0 <> i -> eta((TVar 0) {i := u}) = eta(TVar 0).
Proof.
  simpl.
  intros.
  case (eq_nat_dec i 0).
  intros; omega.
  auto.
Qed.  
  
Lemma subst_var_eta :
  forall u,forall n i,
    i <> n -> eta((TVar n) {i := u}) = eta (TVar n).
Proof.
  induction n.
  intros.
  rewrite (subst_var_0_eta);  auto.
  simpl.
  intros.
  case (eq_nat_dec i (S n)).
  intros.
  omega.
  auto.
Qed.


(** * La propriété IE *)

Theorem IEProp : forall  t u:term,
     sn u /\ sn (t {0 := u}) -> sn(t [u]).
Proof.
  intros. destruct H.
  (* On applique l'induction forte *)
  induction t,u using sn_ind.
  (* On construit nos hypotheses *)
  -apply Ax1. intros.
  (* 
   On enumeres les reduits de t
   par une invertion sur les constructeurs de la reduction 
   *)
   revert H2. inv.
  (* Le cas db se resout par une invertion *)
   *revert H3. inv.  
   (*  cas ls *)
   *revert H3. inv. apply lsa_red in H7. apply sn_and_eta in H0. destruct H0.
    apply H1.
   +unfold alpha. unfold f_cp. apply lt_G. simpl. unfold f_cp.
    assert (eta (u1) {0 := u2} = eta (u1) {0 := shift u2});auto.
    rewrite <- H7 in H4. rewrite H4. f_equal. apply eta_subst_eq. simpl.
    apply (One_step_ls u1 t'0 u2 H3).
   +auto.
   +apply renommage_subst. rewrite H7. auto.

  (*  cas gc *)
    *revert H3. inv.
   +rewrite subst_nothing in H0. apply <- renommage ;auto.


  (*  cas contexte 1  *)
    *apply H1.
   +unfold alpha. unfold f_cp. apply lt_L. simpl. assert (u1 --> t'0). auto.
    apply (lemme2 u1 t'0 u2) in H6. apply Ax2_bisl;auto.
   +auto.
   +apply (Th2 (u1) {0 := u2} (t'0) {0 := u2}). auto. apply lemme2. auto.


  (*  Case contexte 2  *)
    *case (eq_nat_dec  (num_occ u1 0)  0). intros. apply H1.
   +intros. unfold alpha. unfold f_cp. apply lt_L. simpl.
  
  (* case |t|_x = 0 *)
    intros. apply (no_occ_of_x u1 u' u2 0) in e.  rewrite e. 
    assert (
        eta u' < eta u2 ->
        eta (u1) {0 := u2} + eta u' < eta (u1) {0 := u2} + eta u2
      )  by omega. apply H3. apply Th3;auto.
   +apply (Th2 u2 u');auto.
   +apply (no_occ_of_x u1 u' u2 0) in e.  rewrite e. auto.
  
  (* case |t|_x > 0 *)

   +intros. apply H1. unfold alpha. unfold f_cp. apply lt_L. simpl.
    assert (u2 --> u');auto. apply (lemme1 u1 0 u2 u') in H6. apply Ax2_stars_eta in H6.
    assert (
        eta u' < eta u2 -> 
        eta (u1) {0 := u'} <= eta (u1) {0 := u2} -> 
        eta (u1) {0 := u'} + eta u' < eta (u1) {0 := u2} + eta u2
      ) by omega. apply H4. apply (Th3 u2 u');auto. auto. auto.

    apply (Th2 u2 u');auto.
  
    apply (Ax2_stars_sn (u1) {0 := u2} (u1) {0 := u'});auto.
    apply (lemme1 u1 0 u2 u');auto.
Qed.  
  
  