(**
This file is part of the IE library

Copyright (C) El Ouraoui, Mayero, Mazza

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
COPYING file for more details.
*)

Require Export reduction.
Require Import Coq.Program.Equality.

Inductive lsa_d_ret (u :term) (i:nat):  term->term->Prop:= 
| LSA_Var1 :  
    lsa_d_ret u i (TVar i) (lift (S i) u)
| LSA_Lam1 :  forall t t':term,
                lsa_d_ret (u)  (S i) t t' -> 
                lsa_d_ret  u i (\t) (\t')
| LSA_AppLeft1 : forall t t' s:term, 
                   lsa_d_ret u i t t' -> 
                   lsa_d_ret  u i (t @ s) (t' @ s)
| LSA_AppRight1 : forall t t' s:term,
                    lsa_d_ret u i t t' -> 
                    lsa_d_ret u i (s @ t) (s @ t')
| LSA_SubExt1 :  forall t t' s:term,
                   lsa_d_ret u i t t' -> 
                   lsa_d_ret u i (s[t]) (s[t'])
| LSA_SubInt1 : forall t t' s:term, 
                  lsa_d_ret  (u)  (S i) t t' -> 
                  lsa_d_ret  u i (t[s]) (t'[s]).

Hint Resolve LSA_Var1 LSA_Lam1 LSA_AppLeft1 LSA_AppRight1 LSA_SubExt1 LSA_SubInt1 :IE.


(** Autre définition de ls qui retarde la réindexation des 
variables libres *)

Inductive lsa_ret (i:nat) : term -> term -> Prop :=
| LSA_RET : forall t t' u,
              lsa_d_ret u i t t' -> lsa_ret i (t[u]) (t'[u]).

(** Equivalence entre les deux ls *)

Lemma ls_equivalence: forall t t' u,forall i,
             lsa_def (lift i u) i t t' <-> lsa_d_ret u i t t'.
Proof.  
  intros t t' u i.
  split;intro H.
  dependent induction H.  (*try (constructor;auto).*)
  rewrite <- simpl_lift.
  apply LSA_Var1.
  constructor.
  apply IHlsa_def.
  rewrite  <- simpl_lift;auto.
  constructor;auto.
  constructor;auto.
  constructor;auto.
  constructor.
  apply IHlsa_def.
  rewrite  <- simpl_lift;auto.
  induction H;simpl. 
  rewrite simpl_lift.
  apply LSA_Var.
  constructor;auto.
  rewrite <- simpl_lift.
  unfold lift;auto.
  constructor;auto.
  constructor;auto.
  constructor;auto.
  constructor;auto.
  rewrite <- simpl_lift.
  unfold lift;auto.
Qed.

Lemma lsa_Hyp1:
  forall t t' u,forall i, lsa_ret i (t[u]) (t'[u]) <-> lsa_d_ret u i t t'.
Proof.
  intros.
  split.
  intro.
  inversion H. auto.
  apply LSA_RET.
Qed.


Lemma no_subst_gen: forall t u :term, forall i j k:nat,
              j <= k < i+j ->
              (lift_rec (i) t j){k := u} =
              (lift_rec (i) t j).
Proof.  
  induction t.
  -intros; simpl in H. simpl.  case le_gt_dec.
   *simpl. case_eq (eq_nat_dec (k) ( ((i + n))));auto.
    intros. omega.  
   *intros. simpl. case_eq (eq_nat_dec (k) n).
    intros. omega. auto.    
  -intros. simpl. f_equal.
   rewrite IHt. auto.
   omega.
  -intros.  simpl. rewrite IHt1. rewrite IHt2. auto. omega. omega.   
  -intros.  simpl. rewrite IHt1. rewrite IHt2. auto.
   auto. omega.
Qed.



Lemma lsa_red' :
  forall t t' u :term, forall i,
    lsa_d_ret  u i t t' ->
    t'{i:= (lift (S i) u)} = t {i:= (lift (S i) u)}.
Proof.  
  intros t t' u i H.
  induction H;simpl.
  unfold lift.
  rewrite no_subst_gen.
  simpl.
  case_eq (eq_nat_dec i i).
  intros;auto.
  intros;omega.
  omega.


  rewrite <-simpl_lift.
  f_equal.
  rewrite IHlsa_d_ret.
  auto.
  rewrite IHlsa_d_ret.
  auto.
  rewrite IHlsa_d_ret.
  auto.
  rewrite IHlsa_d_ret.
  auto.

  rewrite <-simpl_lift.
  f_equal.
  rewrite IHlsa_d_ret.
  auto.
Qed.

Lemma lsa_red :
  forall t t' u :term,
    lsa_def u 0 t t' ->
    t'{0:= (shift u)} = t {0:= (shift u)}.
Proof.
  -intros. rewrite <- lift0 in H at 1. apply ls_equivalence in H.
   unfold shift. replace (1) with (S 0) by omega. apply lsa_red'. auto.
Qed.