(**
This file is part of the IE library

Copyright (C) El Ouraoui, Mayero, Mazza

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
COPYING file for more details.
*)

Require Export list_subst.
Require Export Decidable.

(** * Règle de réduction db *)

Inductive db : term -> term -> Prop :=
  | DB_Lam : forall t u:term, db ((\t)@u) (t[u])
  | DB_App : forall t u r s, db (t@u) r ->
                                  db ((shift t [s])@u) (shift r [s]).
Hint Resolve DB_Lam DB_App:IE.
Notation " t -->db t'" := (db t t') (at level 66).

(** Lemme de db partiel de lemme 2 de l'article*)

Lemma db_to_subst : forall t t' a x,
                      t -->db t' -> (t {x := a}) -->db (t' {x := a}).
Proof.
  intros.
  induction H;auto with IE.
  simpl.
  auto with IE. 
  simpl in IHdb.
  simpl.
  eapply (DB_App (t {x := a}) (u {x := a})
                 (r {x := a}) (s{x := a})) in IHdb.
  rewrite commut_lift_subst .
  rewrite commut_lift_subst .
  auto.
Qed.

(** * Réduction ls définissant la substitution linéaire *)

Inductive lsa_def (u :term):  nat ->term->term->Prop:= 
  | LSA_Var :forall i,
      lsa_def u i (TVar i) (shift u)
  | LSA_Lam :  forall t t':term,forall i :nat,
                 lsa_def  (shift u) (S i) t t' -> 
                 lsa_def  u i (\t) (\t')
  | LSA_AppLeft : forall t t' s:term,forall i:nat,
                    lsa_def u i t t' -> 
                    lsa_def  u i (t@s) (t'@s)
  | LSA_AppRight : forall t t' s:term,forall i:nat,
                     lsa_def u i t t' -> 
                     lsa_def u i (s@t) (s@t')
  | LSA_SubExt :  forall t t' s:term,forall i:nat,
                  lsa_def u i t t' -> 
                  lsa_def u i (s[t]) (s[t'])
  | LSA_SubInt : forall t t' s:term,forall i:nat,
                  lsa_def (shift u) (S i) t t' -> 
                  lsa_def  u i (t[s]) (t'[s]).

Hint Resolve LSA_Var LSA_Lam LSA_AppLeft LSA_AppRight LSA_SubExt LSA_SubInt :IE.


(** Définition de ls *)
Inductive lsa (i:nat) : term -> term -> Prop :=
| LSA : forall t t' u,  lsa_def u i t t'  -> lsa i (t[u]) (t'[u]).


(*
Un hack permettant de passer à la forme décomposé 
remarque :on sait que lsa i t[u] t'[u] <-> lsa_def u i t t' 
*)

Lemma lsa_Hyp:
  forall t t' u,forall i, lsa i (t[u]) (t'[u]) <-> lsa_def u i t t'.
intros.
split.
intro.
inversion H. auto.
apply LSA.
Qed.

(** lemme de ls partiel de lemme 2 de l'article*)
Lemma lsa_to_subst :
  forall t t',
  forall j:nat,
    lsa 0 t t' ->
    forall a :term,
      lsa 0 t {(0+j) := a} t'{(0+j) := a}.
Proof.


  (*cas variables *)
  intros t1 t2 j H.
  induction H.
  induction H.
  
  intros.
  simpl.
  
  (* pour eliminer le matching par cas utiliser case sur i0 *)
  case i.
  rewrite commut_lift_subst .
  constructor.
  constructor.
  
  intros.
  elim (eq_nat_dec (S n + j) n).
  intros.
  contradict a0.
  omega.


  intros.
  simpl.
  rewrite commut_lift_subst .
  constructor.
  constructor.

    (* cas abstraction *)
  intros.
  simpl.
  constructor.
  apply LSA_Lam.
  
  simpl in IHlsa_def.
  rewrite <- commut_lift_subst.
  apply (lsa_Hyp ( t{ (S (S (i + j))) := shift (shift a)} )
                 ( t'{(S (S (i + j))) := shift (shift a)} )
                 ( (shift u){(S (i + j)) := shift a} ) (S i) ) in IHlsa_def.
  auto with IE.
  
  (* cas applications Left *)
  intros.
  simpl in IHlsa_def .
  simpl.
  constructor.
  apply (lsa_Hyp ( t {(S (i + j)) := shift a})
                 ( t'{(S (i + j)) := shift a} )
                 ( u {(i + j) := a} ) i ) in IHlsa_def.
  auto with IE.
  
  (* cas applications Right *)
  intros.
  simpl in IHlsa_def .
  simpl.
  constructor.
  apply (lsa_Hyp ( t {(S (i + j)) := (shift a) })
                 ( t'{(S (i + j)) := (shift a) } )
                 ( u {(i + j) :=  a }  ) i ) in IHlsa_def.
  auto with IE.
  
  (* cas subst box ext*)
  intros.
  simpl in IHlsa_def .
  simpl.
  constructor.
  apply (LSA_SubExt).
  apply (lsa_Hyp ( t {(S (i + j)) := shift a})
                 ( t'{(S (i + j)) := shift a} )
                 ( u {(i + j) := a} ) i ) in IHlsa_def.

  auto with IE.
  
  (* cas subst box int*)
  intros.
  simpl in IHlsa_def .
  simpl.
  constructor.
  eapply LSA_SubInt.
  rewrite <- commut_lift_subst.
  apply (lsa_Hyp ( t  {(S (S (i + j))) := (shift (shift a))} )
                 ( t' {(S (S (i + j))) := (shift (shift a))} )
                 ( (shift u) {(S (i + j)) := shift a}  ) (S i) ) in IHlsa_def.
  auto with IE.
Qed.

Definition ls := lsa 0 .
Notation " t -->ls t'" := (ls t t') (at level 66).

Lemma ls_to_subst : forall t t' ,forall j:nat, t -->ls t' -> forall a,
                                               t { j := a } -->ls t' { j := a }.
Proof.
  unfold ls.
  apply lsa_to_subst.
Qed.

(** * Règle gc *)

Inductive gc : term -> term -> Prop :=
| GC_Term : forall t u:term,
           gc ((shift t)[u]) t. 
Hint Resolve GC_Term: IE.
Notation " t -->gc t'" := (gc t t') (at level 66).


(** Lemme de gc partiel du lemme 2 de l'article*)

Lemma gc_to_subst : forall t t' a, t -->gc t' ->forall k,
                                   (t{k := a}) -->gc (t'{k := a}).
Proof.
  induction 1.
  simpl.
  intros.
  rewrite (commut_lift_subst).
  auto with IE.
Qed.



(** * Réduction complète  *)

Inductive red : term -> term -> Prop :=
  | DB : forall t t':term, t  -->db t' -> red t t'
  | LS : forall t t':term, t  -->ls t' -> red t t'
  | GC : forall t t':term, t  -->gc t' -> red t t'
                                                              
  | RedContextTAbs:
      forall t t':term,
      red t t' ->
      red (\t) (\t')
  | RedContextTAppLeft:
      forall t t' v:term,
      red t t' ->
      red (t @ v) (t' @ v)
  | RedContextTAppRight:
      forall v v' t:term,
      red v v' ->
      red (t @ v) (t @ v')
  | RedContextSubExt :
      forall t t' u:term,
      red t t' ->
      red (t[u]) (t'[u])
  | RedContextSubInt :
      forall t u u':term,
      red u u' ->
      red (t[u]) (t[u']).  

Notation " t --> t'" := (red t t') (at level 66).

(** Réduction star *)

Inductive red_star: term -> term -> Prop :=
  | red_star_eq: forall t:term, red_star t t
  | red_star_step: forall t u v:term,
                     red t u -> red_star u v -> red_star t v.

Notation " t -->* t'" := (red_star t t') (at level 66).

Hint Resolve DB LS GC RedContextTAbs RedContextTAppLeft RedContextTAppRight RedContextSubExt GC_Term: IE.

Theorem red_to_star:
  forall t u:term, (t --> u) -> (t -->* u).
Proof.
  intros.
  apply (red_star_step t u u).
  trivial.
  apply red_star_eq.
Save.


Theorem red_star_db:
  forall t t': term, t -->db t' -> t -->* t'.
Proof.
  intros.
  apply (red_to_star t t').
  apply DB.
  trivial.
Save.

Theorem red_star_ls:
  forall t t': term,
    t -->ls t' -> t -->* t'.
Proof.
  intros.
  apply red_to_star.
  apply (LS t t').
  trivial.
Save.

Theorem red_star_gc:
  forall t t': term,
    t -->gc t' -> t -->* t'.
Proof.
  intros.
  apply red_to_star.
  apply (GC t t').
  trivial.
Save.


Theorem red_star_ContextTAbs:
  forall t t': term, t -->* t' -> (\t) -->* (\t').
Proof.
  intros t t' H; induction H.
  apply red_star_eq.
  apply (red_star_step (\t) (\u) (\v)); trivial.
  apply RedContextTAbs; trivial.
Save.

Theorem red_star_RedContextTAppLeft:
  forall t t' u: term, t -->* t' -> (t @ u) -->* (t' @ u).
  intros t1 t2 v H; induction H.
  apply red_star_eq.
  apply (red_star_step (t @ v) (u @ v) (v0 @ v)); trivial.
  apply RedContextTAppLeft; trivial.
Save.

Theorem red_star_RedContextTAppRight:
  forall w w' t: term, w -->* w' -> (t @ w) -->* (t @ w').
  intros t u v H; induction H.
  apply red_star_eq.
  apply (red_star_step (v @ t) (v @ u) (v @ v0)); trivial.
  apply RedContextTAppRight; trivial.
Save.

Theorem red_star_RedContextSubExt:
  forall t t' u: term, t -->* t' -> (t[u]) -->* (t'[u]).
Proof.
  intros;induction H.
  apply red_star_eq.
  apply (red_star_step (t[u]) (u0[u]) (v[u])); trivial.
  apply RedContextSubExt; trivial.
Save.

Theorem red_star_RedContextSubInt:
  forall w w' u:term, w -->* w' -> (u[w]) -->* (u[w']).
Proof.
  intros;induction H.
  apply red_star_eq.
  apply (red_star_step  (u[t]) (u[u0]) (u[v])); trivial.
  apply RedContextSubInt; trivial.
Save.

  
Lemma red_app : forall t1 t1' t2 t2' ,
                  t1 -->* t1' -> t2 -->* t2'  -> (t1 @ t2) -->* (t1' @ t2').
Proof.    
  simple induction 1.
  simple induction 1.
  intros.
  apply red_star_RedContextTAppRight.
  apply red_star_eq.
  intros.
  apply red_star_RedContextTAppRight.
  apply (red_star_step t0 u v);auto.
  intros.
  apply red_star_step with (u @ t2).
  apply RedContextTAppLeft;auto.
  auto.
Qed.

Lemma red_sub : forall t1 t1' t2 t2' ,
                  t1 -->* t1' -> t2 -->* t2'  -> (t1[t2]) -->* (t1'[t2']).
Proof.    
  simple induction 1.
  simple induction 1.
  intros.
  apply red_star_RedContextSubExt.
  apply red_star_eq.
  intros.
  apply red_star_RedContextSubInt.
  apply (red_star_step t0 u v);auto.
  intros.
  apply red_star_step with (u[t2]).
  apply RedContextSubExt;auto.
  auto.
Qed.

(**  * Le lemme 2 de l'article : si t -> t' => t{k/u} -> t'{k/u} 
*)


Lemma lemme2' : forall t t', t --> t' -> forall u k, t{k := u} --> t'{k := u}.

Proof.
  simple induction 1;simpl;auto with IE.
  intros.
  apply DB.
  eapply (db_to_subst t0 t'0 u);auto.
  intros.
  eapply (LS (t0{k := u}) (t'0{k := u})).
  eapply (ls_to_subst t0 t'0 k H0);auto.
  intros.
  eapply GC .
  apply (gc_to_subst t0 t'0 u H0 k );auto.
  intros.
  apply RedContextSubInt.
  auto.
Qed.
Lemma lemme2 : forall t t' u, t --> t' -> t{0 := u} --> t'{0 := u}.
Proof.
  intros. eapply (lemme2' t t' H u 0).
Qed.


Lemma db_to_list_substT :
  forall u u' j k,
    u -->db u' -> (list_substT u (build_list (fv_rec u k) j) k j) -->db (list_substT u' (build_list (fv_rec u' k) j) k j).
Proof.
  intros u u' j k H.
  induction H;auto with IE.
  simpl.
  auto with IE.
  simpl in IHdb.
  simpl.
  eapply (DB_App (list_substT t (build_list (fv_rec t k) j) k j) (list_substT u (build_list (fv_rec u k) j) k j) (list_substT r (build_list (fv_rec r k)  j) k j) (list_substT s (build_list (fv_rec s k) j) k j)) in IHdb.
  simpl in IHdb.
  rewrite  commut_lift_subst_list;auto.
  rewrite  commut_lift_subst_list;auto.
Qed.

Lemma lsa_to_list_substT :
  forall   t1 t2 j k,
    lsa 0 t1 t2 -> lsa 0 (list_substT t1 (build_list (fv_rec t1 (0+k)) j) (0+k) j) (list_substT t2 (build_list (fv_rec t2 (0+k)) j) (0+k) j).
Proof.
  intros t1 t2 j k H.
  induction H.
  induction H;simpl.
  (* case variable *)
  constructor.
  case i.
  rewrite  commut_lift_subst_list;auto.
  constructor.
  
  intros.
  elim (le_gt_dec (S n + k) n).
  intros.
  contradict a.
  omega.
  
  intros.
  simpl.
  rewrite  commut_lift_subst_list;auto.
  constructor.
  
  (* cas abstraction *)
  
  intros.
  constructor.
  simpl.
  eapply LSA_Lam.
  apply lsa_Hyp in IHlsa_def.
 
  simpl in IHlsa_def.
  rewrite  <- commut_lift_subst_list;auto.

   
  (* cas applications Left *)
  
  intros.
  constructor.
  apply lsa_Hyp in IHlsa_def.
  auto with IE.


  (* cas applications Right *)
  
  intros.
  constructor.
  apply lsa_Hyp in IHlsa_def.
  auto with IE.

  (* cas subst box ext*)
  
  intros.
  constructor.
  apply lsa_Hyp in IHlsa_def.
  apply (LSA_SubExt).
  auto with IE.

  (* cas subst box int*)
  
  intros.
  constructor.
  apply lsa_Hyp in IHlsa_def.
  eapply LSA_SubInt.
  rewrite  <- commut_lift_subst_list;auto.
Qed.


Lemma ls_to_list_subst :
  forall t1 t2 j,forall k:nat, t1 -->ls t2 ->
                               (list_substT t1 (build_list (fv_rec t1 k) j) (k) j) -->ls
                               (list_substT t2 (build_list (fv_rec t2 k) j) (k) j).
Proof.
  unfold ls.
  apply (lsa_to_list_substT ).
Qed.

Lemma gc_to_list_substT : forall t1 t2, t1 -->gc t2 ->forall k j,
                          (list_substT t1 (build_list (fv_rec t1 k) j) k j) -->gc
                          (list_substT t2 (build_list (fv_rec t2 k) j) k j).
Proof.  
  induction 1.
  simpl.
  intros.
  rewrite   commut_lift_subst_list .
  auto with IE.
Qed.

(*  Lemme intermédiaire: 
fv u = {k0...kn} et fv u' = {k'0...k'n}
  si u -> u' => t{k0+k/k0}...{kn+k/kn} -> t{k'0+k/k'0}...{k'n+k/k'n} 
 *)

Lemma red_to_list_substT :
  forall  j u u', u --> u' ->
                  forall k, (list_substT u (build_list (fv_rec u k) j) k j) -->
                            (list_substT u' (build_list (fv_rec u' k) j) k j).
Proof.
  simple induction 1;simpl;auto with IE.
  intros.
  apply DB.
  eapply (db_to_list_substT t t');auto.
  intros.
  eapply (LS (list_substT t (build_list (fv_rec t k) j) k j)
             (list_substT t' (build_list (fv_rec t' k) j) k j)
         ).
  eapply (ls_to_list_subst t t' j k H0);auto.
  intros.
  eapply GC .
  apply (gc_to_list_substT t t' H0 k j );auto.
  intros.
  apply RedContextSubInt.
  auto.
Qed.


(**  * Le lemme 1 de l'article:  si u -> u' => t{k/u} -> t{k/u'} 
 *)

Lemma lemme1 : forall t x u u', u --> u' -> t{x := u} -->* t {x := u'}.
Proof.
  induction t.
  intros.  
  (* cas variables *)
  simpl.
  case (eq_nat_dec x n).
  intros.
  apply red_to_star;auto.
  intros;apply red_star_eq;auto.

  (* cas abstaction *)
  intros.
  simpl.
  
  apply (red_star_ContextTAbs (t{S x := shift u}) (t {S x := shift u'})) in IHt.
  auto.

  rewrite (shift_subst u 1 0  (fv u));auto with IE.
  rewrite (shift_subst u' 1 0 (fv u'));auto with IE.

  apply (red_to_list_substT 1 u u' H 0).

 
  (* cas application *)
  intros.
  simpl.
  apply red_app;auto.

  (* cas substitution *)
  intros.
  simpl.
  apply red_sub;auto.
  apply IHt1.
  rewrite (shift_subst u 1 0  (fv u));auto with IE.
  rewrite (shift_subst u' 1 0 (fv u'));auto with IE.
  apply (red_to_list_substT 1 u u' H 0).
Qed.
